;===============================================================================
; Алиасы, 
;===============================================================================

; Индексные регистры:
.alias SP 39            ; Указатель вершины стека.
.alias BP 38            ; Указателя базы функции (фрейм).
.alias SI 37            ; Адрес источника данных.
.alias DI 36            ; Адрес приемника данных.
.alias CX 35            ; Счетчик.


; NULL
.global nullMCP__
.data
nullMCP__:
	.long 0



; Константы системные:
.global sizeof.byte     ; Размер байт
.equiv  sizeof.byte,      1
.global sizeof.short    ; Размер короткого целого
.equiv  sizeof.short,     2
.global sizeof.long     ; Размер длинного целого
.equiv  sizeof.long,      4
.global sizeof.quad     ; Размер 64-х битного
.equiv  sizeof.quad,      8
.global sizeof.ptr      ; Размер указателя (адреса памяти).
.equiv  sizeof.ptr,       sizeof.long
.global sizeof.DM       ; Размер памяти данных.
.equiv  sizeof.DM,        0x20000

.global PSW.ONIRQS      ; Признак обработки немаскируемых прерываний
.equiv  PSW.ONIRQS,       0x0001
.global INTR.PPGE       ; Немаскируемое программное исключение
.equiv  INTR.PPGE,        0x0008

.alias divsl divsl.crt
.alias divl divl.crt

; Глобальные имена:
.global irq.desc.tbl    ; Таблица векторов прерываний.
.global sizeof.IDT.item ; Размер элемента таблицы векторов прерываний.
.equiv  sizeof.IDT.item,  sizeof.ptr
.global __libc_main     ; Ф-ция инициализации стандартной библиотеки Си.
.global main            ; Главная ф-ция задачи.
.global divsl           ; Деление знаковых чисел.
.global divl            ; Деление беззнаковых чисел.
.global movb            ; Копирование участком памяти байтами.
.global movl            ; Копирование участком памяти словами.
.global movq            ; Копирование участком памяти двойными словами.
.global master.isr.retaddr

//.weak sysTimerHandler	; Обработчик системного таймера по умолчанию
.global initSysTimer	; Инициализация системного таймера.

; Локальные имена:
.local master.isr       ; Первичный обработчик прерываний.
.local tsk.retval       ; Код завершения задачи.


.global iSP
.global iBP
.global gSP
.global gBP
.data
	iSP:
		.long 0
	iBP:
		.long 0
	gSP: 
		.long 0
	gBP:
		.long 0


;===============================================================================
; Interrupt Descriptor Table (IDT): Таблица векторов прерываний
;===============================================================================
.data

  irq.desc.tbl:       
    .long \
      INMI.isr,\
      ENMI.isr,\
      PERE.isr,\
      PPGE.isr,\
      MPRGE.isr,\
      SWT.isr,\
      SWI.isr,\
      UART0.isr, UART1.isr, UART2.isr, UART3.isr,\
      I2C0.isr, I2C1.isr,\
      SPI0.isr, SPI1.isr, SPI2.isr,\
      I2S0.isr,\
      GPTIM0.isr, GPTIM1.isr, GPTIM2.isr, GPTIM3.isr, GPTIM4.isr,\
      GPTIM5.isr, GPTIM6.isr,\
      PWM0.isr,\
      RTC.isr,\
      GPIOA.isr, GPIOB.isr, GPIOC.isr, GPIOD.isr,\
      ETHERNET0.isr,\
      USB0.isr

;===============================================================================
; Начальный код
;-------------------------------------------------------------------------------
; - инициализирует стеки
; - утанавливает адрес первичного обработчика прерываний
; - разрешает обработку прерываний
; - уходит на функцию __libc_main определенную в стандартной библиотеке
;   которая в свою очередь вызывает ф-цию main программы.
; - записывает возвращаемое значение в переменную tsk.retval
; - инициирует немаскируемое программное прерывание.
;===============================================================================
.data
  tsk.retval: .long               ; Результат выполнения задачи.
  .alias sizeof.tsk.retval sizeof.long

.text

  start:
    getl 0x00000000
    setl #MODR, @1

    ; Инициализация стека и указателей стека:
    ;     0           #SP      #BP = sizeof.DM
    ;     |           |        |
    ;     v           v        v
    ; DM |...........|....|....|
    ;                      ^^^^
    ;                 ^^^^ tsk.retval
    ;                 #start.stop
    ; #SP -- адрес возврата из программы;
    ; #BP -- за верхнюю границу DM.

    jmp   __libc_main
    getl  sizeof.DM
    getl  0
    getl  start.stop
    subl  @3, sizeof.tsk.retval
    subl  @1, sizeof.ptr
    wrl   @4, @2                  ; __libc_main::retval = 0
    wrl   @4, @2                  ; __libc_main::retaddr = start.stop
    setl  #BP, @7                 ; BP = sizeof (DM)
    setl  #SP, @4                 ; SP = BP - sizeof.tsk.retval - sizeof.ptr
    
    ; Инициализация обработки прерываний:

    setl  #IHOOKADDR, master.isr  
    getq  #PSW
    or    @1, PSW.ONIRQS
    setq  #PSW, @1

  complete                        ; => __libc_main
  start.stop:                     ; <= __libc_main

    jmp   start.hook

    ; Копируем возвращаемое значение во статическую переменную.

    rdl   #SP, sizeof.ptr
    wrl   @1, tsk.retval

    ; Освобождаем память в стеке.

    getl  #SP
    addl  @1, sizeof.ptr + sizeof.tsk.retval
    setl  #SP, @1                 ; SP = BP = sizeof.DM

    ; Устанавливаем немаскируемое программное прерывание

    getq  #INTR
    or    @1, INTR.PPGE
    setq  #INTR, @1

  complete
  start.hook:

    jmp   start

    ; Несмотря на то, что в предыдущем параграфе выставляется бит
    ; немаскируемого программного прерывания, управление будет передано
    ; именно на этот параграф. И только по его окончанию будет выполнен
    ; уход на прерывание, при этом в качестве адреса возврата из прерывания
    ; будет сохранен адрес, выставляемый в этом параграфе.
    ; Переход на начальный адрес позволит обработчику немаскируемого
    ; программного прерывания, вызванного окончанием программы либо
    ; завершить выполнение (что и делается сейчас в заглушке обработчика)
    ; либо зациклить выполнение программы, штатным образом обработав
    ; возврат из обработчика прерывания.

  complete

;===============================================================================
; __libc_main
;-------------------------------------------------------------------------------
; Заглушка для ф-ции которая должна предоставляться стандартной библиотекой Си.
; Выполняет инициализацию окружения выполнения программы на Си и вызывает
; фцнкцию main.
;===============================================================================
  .alias __libc_main.ret.value  #SP,sizeof.ptr
  
  __libc_main:

    jmp   main                ; Транслируем управление ф-ции main

  complete
  
;===============================================================================
; Interrupt Service Routine (ISR): Обработчик прерываний
;===============================================================================
; Нереентерабельная обработка прерываний.

.data

  master.isr.retaddr: .long   ; Адрес возврата из прерывания.

.text

master.isr:

	jmp mi.hendler

	saveSP := getl #SP
	saveBP := getl #BP
	loadSP := rdl gSP
	loadBP := rdl gBP
	reta := getl  #IRETADDR

	wrl @saveSP, iSP
	wrl @saveBP, iBP
	setl #SP, @loadSP
	setl #BP, @loadBP
	wrl   @reta, master.isr.retaddr

complete

mi.hendler:
    ; Формируем переход на обработчик прерывания
    ; по адресу из таблицы прерываний.
    ; Стек не трогается т.к. в нем могут передаваться аргументы для
    ; программного прерывания, за исключением того, что добавляем в стек
    ; адрес возврата на окончание первичного обработчика прерывания.
  
    getl  #SP
    getl  mi.stop
    getl  #INTNUMR
    getl  irq.desc.tbl
    mull @2, sizeof.IDT.item
    addl @1, @2
    rdl   @1
    jmp   @1
    subl  @8, sizeof.ptr
    wrl   @8, @1
    setl  #SP, @2

    ; Сохраняем адрес возврата из прерывания.

  complete

mi.stop:
	
	jmp mi.PF

	getl #SP
	addl @1, sizeof.ptr
	setl #SP, @1

complete


mi.PF:

    ; Возвращаемся из прерывания:

	rdl   master.isr.retaddr
	jmp   @1

	; Разрешаем прерывания:

	getq  #PSW
	or    @1, PSW.ONIRQS
	setq  #PSW, @1

	saveSP := getl #SP
	saveBP := getl #BP
	loadSP := rdl iSP
	loadBP := rdl iBP

	wrl @saveSP, gSP
	wrl @saveBP, gBP
	setl #SP, @loadSP
	setl #BP, @loadBP
	
complete

  STUB.isr:

    rdl   #SP
    jmp   @1

  complete

  INMI.isr:
  ENMI.isr:
  PERE.isr:
  PPGE.isr:

    getl  0

  complete                    ; Переход в никуда.

  MPRGE.isr:
  SWI.isr:
  UART0.isr: UART1.isr: UART2.isr: UART3.isr:
  I2C0.isr: I2C1.isr:
  SPI0.isr: SPI1.isr: SPI2.isr:
  I2S0.isr:
  GPTIM0.isr: GPTIM1.isr: GPTIM2.isr: GPTIM3.isr: GPTIM4.isr: GPTIM5.isr: GPTIM6.isr:
  PWM0.isr:
  RTC.isr:
  GPIOA.isr: GPIOB.isr: GPIOC.isr: GPIOD.isr:
  ETHERNET0.isr:
  USB0.isr:

    jmp STUB.isr

  complete

;===============================================================================
; Реализация функции, которая инициализирует системный таймер
; Интерфейс вызова ф-ции соответствует следующему объявлению в синтаксисе Си:
;
; void initSysTimer(bool isCycle, unsigned short period, unsigned char prediv);
;
;===============================================================================

.text

  .alias initSysTimer.retAddr	 #SP
  .alias initSysTimer.result	 #SP, +4
  .alias initSysTimer.isCylce	 #SP, +8
  .alias initSysTimer.period	 #SP, +12
  .alias initSysTimer.preDiv	 #SP, +16
  
  initSysTimer:
    rdl initSysTimer.retAddr
    rdl initSysTimer.isCylce
    rdl initSysTimer.period
    rdl initSysTimer.preDiv

    jmp @4
    and @4, 0x00000001
    and @4, 0x0000FFFF
    and @4, 0x000000FF

    getl 0x00000001	// бит 0 STCR - разрешение работы счётчика
    slll @4, 1		// бит 1 STCR - признак циклической работы таймера
    slll @3, 8		// биты 8 - 15 STCR - prediv
    setl #STVALR, @5

    or @4, @3
    or @1, @3
    getl 0x00000000
    getl #MSKR

    or @1, 0x00000002
    wrl @3, initSysTimer.result
    setl #STCR, @5
    setl #MSKR, @3
  complete

;===============================================================================
; Реализация обработчика прерывания от системного таймера
;
; 1. При прерывании от системного таймера первичный обработчик прерывания
;    передаёт управление на обработчик SWT.isr
; 2. В обработчике осуществляется очистка признака завершения счета
;    периода, заданного в STVAL, в циклическом режиме работы
; 3. В обработчике устанавливается адрес возврата после выполнения
;    пользовательского обработчика прерывания sysTimerHandler
; 4. После выполнения данного обработчика управление передаётся
;    на пользовательский обработчик прерывания sysTimerHandler
;    Интерфейс ф-ции соответствует следующему объявлению в синтаксисе Си:
;
;      void sysTimerHandler();
;
;    Если реализация данной функции отсутствует (пользователь не написал
;    код для обработчика прерывания от системного таймера), то выполняется
;    обработчик sysTimerHandler по умолчанию, который определён ниже и
;    имеет значение аттрибута связывания WEAK (смотри раздел "Глобальные
;    имена в начале данного файла").
;===============================================================================
  SWT.isr:
    jmp sysTimerHandler
    getl #SP
    getl SWT.isr.epiloge
    getl #STCR

    subl @3, sizeof.ptr
    and @2, ~0x00000004
    wrl @4, @2
    setl #STCR, @2

    setl #SP, @4
    setl #STVALR, 0xffff

  complete

  
  SWT.isr.epiloge:
    jmp STUB.isr
    getl #SP
    addl @1, sizeof.ptr
    setl #SP, @1

  complete

;===============================================================================
; Реализация обработчика прерывания от системного таймера по умолчанию
;
; Данный обработчик выполняется только в том случае, если не определён
; пользовательский обработчик прерывания (пользователь не написал
; код для обработчика прерывания от системного таймера)
;
; Значение аттрибута связывания данного обработчика (а точнее символа
; sysTimerHandler) устанавливается в WEAK (смотри раздел "Глобальные
; имена в начале данного файла").
; Код данного обработчика просто возвращает управление, не выполняя
; никаких других действий.
;===============================================================================

 // sysTimerHandler:
//	jmp #SP
 // complete


;===============================================================================
; Целочисленное деление знаковых слов.
; (Использует алгоритм беззнакового деления.)
; Если делитель ноль, то результат не определен.
; Интерфейс вызова ф-ции соответствует следующему объявлению в синтаксисе Си:
;
;   union Result {
;     uint64_t value;
;     struct {
;       int32_t Quotient;
;       int32_t Remainder;
;     } fields;
;   }
; Result divsl (int32_t Numerator, int32_t Denominator);
; 
; 1. Изменить знак у делимого если оно отрицательное. Сделать тоже для делителя.
; 2. Преобразовать делимое и делитель в беззнаковое значение.
; 3. Использовать алгоритм беззнакового деления.
; 4. Преобразовать частное и остаток в знаковое представление.
; 5. Изменить знак частного, если знаки делимого и делителя были различные.
; 6. Если делимое отрицательное, изменить знак остатка.
;===============================================================================

.text

  .alias div.Denominator  #BP, +20
  .alias div.Numerator    #BP, +16
  .alias div.Remainder    #BP, +12
  .alias div.Quotient     #BP, +8
  .alias div.RetAddr      #BP, +4
  .alias div.PrevBP       #BP
  .alias div.Index        #BP, -4

  .alias div.arg.Denominator  #SP, -4
  .alias div.arg.Numerator    #SP, -8
  .alias div.arg.Remainder    #SP, -12
  .alias div.arg.Quotient     #SP, -16
  .alias div.arg.RetAddr      #SP, -20

  .alias div.sizeof.args      4*sizeof.long

  .alias div.res.Remainder    #SP, +8
  .alias div.res.Quotient     #SP, +4

  divsl:
  divsl.prologue:

    jmp   divsl.check.numerator

    getl  #SP
    getl  #BP
    subl  @2, sizeof.ptr
    wrl   @2, #SP,-sizeof.ptr
    setl  #SP, @2
    setl  #BP, @3

  complete
  divsl.check.numerator:               ; if (N < 0)

    rdsl   div.Numerator
    js    @1, divsl.neg.num
    jns   @2, divsl.pos.num
  
  complete
  divsl.neg.num:                       ; if (D < 0)

    rdsl   div.Denominator
    js    @1, divsl.neg.num.neg.denom
    jns   @2, divsl.neg.num.pos.denom

  complete
  divsl.neg.num.neg.denom:             ; divl (-N,-D)

    jmp   divl

    getl  #SP

    rdsl  div.Denominator
    rdsl  div.Numerator
    getl  divsl.neg.num.neg.denom.returns

    insubsl @3, 0
    insubsl @3, 0
    subl  @6, div.sizeof.args + sizeof.ptr

    wrsl  @3, div.arg.Denominator
    wrsl  @3, div.arg.Numerator
    wrl   @6, div.arg.RetAddr

    setl  #SP, @4
  
  complete
  divsl.neg.num.neg.denom.returns:      ; R = -R

    jmp   divsl.epilogue

    getl  #SP
    
    rdl   div.res.Remainder
    rdl   div.res.Quotient

    insubsl @2, 0
    addl  @4, div.sizeof.args + sizeof.ptr

    wrl   @2, div.Remainder
    wrl   @4, div.Quotient

    setl  #SP, @3

  complete
  divsl.neg.num.pos.denom:              ; divl (-N,D)

    jmp   divl

    getl  #SP

    rdsl  div.Denominator
    rdsl  div.Numerator
    getl  divsl.neg.num.pos.denom.returns

    insubsl @2, 0
    subl  @5, div.sizeof.args + sizeof.ptr

    wrsl  @5, div.arg.Denominator
    wrsl  @3, div.arg.Numerator
    wrl   @5, div.arg.RetAddr

    setl  #SP, @4

  complete
  divsl.neg.num.pos.denom.returns:      ; Q = -Q; R = -R;

    jmp   divsl.epilogue

    getl  #SP
    
    rdl   div.res.Remainder
    rdl   div.res.Quotient

    insubsl @2, 0
    insubsl @2, 0
    addl  @5, div.sizeof.args + sizeof.ptr

    wrl   @3, div.Remainder
    wrl   @3, div.Quotient

    setl  #SP, @3

  complete
  divsl.pos.num:                        ; if (D < 0)

    rdsl  div.Denominator
    js    @1, divsl.pos.num.neg.denom
    jns   @2, divsl.pos.num.pos.denom

  complete
  divsl.pos.num.neg.denom:              ; divl (N,-D)

    jmp   divl

    getl  #SP

    rdsl  div.Denominator
    rdsl  div.Numerator
    getl  divsl.pos.num.neg.denom.returns

    insubsl @3, 0
    subl  @5, div.sizeof.args + sizeof.ptr

    wrsl  @2, div.arg.Denominator
    wrsl  @5, div.arg.Numerator
    wrl   @5, div.arg.RetAddr

    setl  #SP, @4

  complete
  divsl.pos.num.neg.denom.returns:      ; Q = -Q

    jmp   divsl.epilogue

    getl  #SP
    
    rdl   div.res.Remainder
    rdl   div.res.Quotient

    insubsl @1, 0
    addl  @4, div.sizeof.args + sizeof.ptr

    wrl   @4, div.Remainder
    wrl   @3, div.Quotient

    setl  #SP, @3

  complete
  divsl.pos.num.pos.denom:               ; divl (N,D)

    jmp   divl

    getl  #SP

    rdsl  div.Denominator
    rdsl  div.Numerator
    getl  divsl.pos.num.pos.denom.returns

    subl  @4, div.sizeof.args + sizeof.ptr

    wrsl  @4, div.arg.Denominator
    wrsl  @4, div.arg.Numerator
    wrl   @4, div.arg.RetAddr

    setl  #SP, @4

  complete
  divsl.pos.num.pos.denom.returns:

    jmp   divsl.epilogue

    getl  #SP
    
    rdl   div.res.Remainder
    rdl   div.res.Quotient

    addl  @3, div.sizeof.args + sizeof.ptr

    wrl   @3, div.Remainder
    wrl   @3, div.Quotient

    setl  #SP, @3

  complete
  divsl.epilogue:

    ; Строим возврат.

    rdl   div.RetAddr
    jmp   @1

    getl  #BP
    rdl   div.PrevBP
    setl  #BP, @1
    addl  @3, sizeof.ptr
    setl  #SP, @1

  complete

;==============================================================================
; Целочисленное деление беззнаковых слов.
; Если делитель ноль, то результат не определен.
; Интерфейс вызова ф-ции соответствует следующему объявлению в синтаксисе Си:
;
;   union Result {
;     uint64_t value;
;     struct {
;       uint32_t Quotient;
;       uint32_t Remainder;
;     } fields;
;   }
; Result divl (uint31_t Numerator, uint31_t Denominator);
;==============================================================================

  divl:
  divl.prologue:

    jmp   divl.arg.check

    getl  #SP
    getl  #BP
    subl  @2, sizeof.ptr
    wrl   @2, #SP,-sizeof.ptr
    setl  #SP, @2
    setl  #BP, @3

  complete
  divl.arg.check:                       ; Делимое должно быть больше делителя.

    rdl   div.Numerator
    rdl   div.Denominator

/*  Из-за аппаратной ошибки выставления флага переноса/заёма
*   этот код для некоторых значений работает неправильно
*    subl  @2, @1                        ; if ( N < D )
*    jb    @1, divl.simple.result
*    jae   @2, divl.prepare
*/

    maxl  @1, @2
    subl  @3, @1                        ; if ( N < D )
    jne    @1, divl.simple.result
    je   @2, divl.prepare

  complete

  divl.simple.result:                   ; Если делимое все таки меньше делителя, то оно и есть остаток.

    jmp   divl.epilogue
    getl  0
    rdl   div.Numerator
    wrl   @2, div.Quotient
    wrl   @2, div.Remainder
  
  complete
  divl.prepare:                         ; Выравниваем делитель таким образом, что знакоместо старшего бита
                                        ; делителя соответствует знакоместу старшего бита делимого.
    jmp   divl.check.remainder
    rdl   div.Numerator
    rdl   div.Denominator
                                        ; За компанию, превращаем делимое в остаток,
                                        ; чем оно и будет являться на протяжении выполнения алгоритма.
    wrl   @2, div.Remainder
    normsq @3                           ; Позиция старшего значащего бита делимого
    normsq @3                           ; Позиция старшего значащего бита делителя,
                                        ; должна быть <= чем у делимого, 
                                        ; т.к. исходим из условия, что делимое > делителя.
    subl  @1, @2                        ; Расстояние на которое нужно сдвинуть делитель
                                        ; чтобы старшие значащие биты делимого и делителя совпадали.
    slll  @5, @1                        ; Сдвигаем делитель таки образом,
                                        ; чтобы старшие значащие биты делимого и делителя
                                        ; совпадали по знакоместу.
    wrl   @1, div.Denominator
                                        ; Устанавливаем бит базы-индекса в позицию старшего значащего бита делителя.
    getl  1                             ; Выставляем бит базы-индекса...
    slll  @1, @4                        ; ... и сдвигаем ее так же как и делитель.
    wrl   @1, div.Index
                                        ; Обнуляем начальное значение частного.
    getl  0
    wrl   @1, div.Quotient

  complete
  divl.check.remainder:                 ; Проверяем условие подбора очередной степени частного.
                                        ; Если в данной позиции (степени) делителя остаток больше делителя, то
                                        ; значит можно вычесть из остатка данную степень делителя и, соответсвенно,
                                        ; скорректировать частное.
                                        ; Иначе -- понижаем степень делителя.
    rdl   div.Remainder
    rdl   div.Denominator

/*  Из-за аппаратной ошибки выставления флага переноса/заёма
*   этот код для некоторых значений работает неправильно
*    subl  @2, @1                        ; if R < D
*    jb    @1, divl.next.order
*    jae   @2, divl.calculate.remainder
*/

    maxl @1, @2
    subl @3, @1
    jne    @1, divl.next.order
    je   @2, divl.calculate.remainder

  complete
  divl.calculate.remainder:             ; Вычисляем остаток и частное.

    jmp   divl.next.order
    rdl   div.Remainder
    rdl   div.Denominator
    subl  @2, @1                        ; Вычисляем остаток:     (R-D').
    wrl   @1, div.Remainder             ; Сохраняем остаток: R <- (R-D').
    
    rdl   div.Quotient                  ; Частное.
    rdl   div.Index
    addl  @2, @1                        ; Складываем частное с базой-индексом: Q'+I
    wrl   @1, div.Quotient              ; Запоминаем частное: Q'<=(Q'+I)
  
  complete
  divl.next.order:                      ; Сдвигаемся к следующему разряду.

    rdl   div.Denominator
    rdl   div.Index
    slrl  @2, 1                         ;     (D'>>1)
    slrl  @2, 1                         ;     (I >>1)
    wrl   @2, div.Denominator           ; D'<=(D'>>1)
    wrl   @2, div.Index                 ; I <=(I >>1)
                                        ; Проверяем что все разряды обработаны.
                                        ; if (I == 0) // База-индекс в младщем разряде.
    je    @3, divl.epilogue             ; then Алгоритм завершен
    jne   @4, divl.check.remainder
  
  complete
  divl.epilogue:

    ; Строим возврат.

    rdl   div.RetAddr
    jmp   @1

    getl  #BP
    rdl   div.PrevBP
    setl  #BP, @1
    addl  @3, sizeof.ptr
    setl  #SP, @1

  complete

;==============================================================================
; Копирование памяти.
; Интерфейс вызова ф-ции соответствует следующему объявлению в синтаксисе Си:
;   void movb (void * SrcAddr, void * DstAddr, uint32_t RepeatNumber);
;   void movl (void * SrcAddr, void * DstAddr, uint32_t RepeatNumber);
;   void movq (void * SrcAddr, void * DstAddr, uint32_t RepeatNumber);
; Функции отличаются размером единицы копирования данных.
;==============================================================================

.text

  .alias mov.Counter  #BP, +16
  .alias mov.DesAddr  #BP, +12
  .alias mov.SrcAddr  #BP, +8
  .alias mov.RetAddr  #BP, +4
  .alias mov.PrevBP   #BP

  .alias MODR.SI SI-32
  .alias MODR.DI DI-32
  .alias MODR.CX CX-32

  movb:
  movb.prologue:

    jmp   movb.shadowing

    getl  #SP
    getl  #BP
    subl  @2, sizeof.ptr
    wrl   @2, #SP,-sizeof.ptr
    setl  #SP, @2
    setl  #BP, @3

  complete
  movb.shadowing:

    jmp   movb.loop

    ;=== Вычисляем тени регистра счетчика ===
    ; Вычисляем маску для заданного количества итераций,
    ; таким образом, что в ней выставлены все биты до бита представляющего
    ; наименьшую степень 2 не уступающую количеству итераций:

    rdl   mov.Counter
    subl  @1, 1               ; M - 1
    slrl  @1, 1
    or    @2, @1              ; M = M | (M >> 1)
    slrl  @1, 2
    or    @2, @1              ; M = M | (M >> 2)
    slrl  @1, 4
    or    @2, @1              ; M = M | (M >> 4)
    slrl  @1, 8
    or    @2, @1              ; M = M | (M >> 8)
    slrl  @1, 16
    or    @2, @1              ; Mask = M | (M >> 16)
    
    ; Расчитанная маска позволит выполнить количество итераций, соотвтетсвующее
    ; ближайшей сверху степени двойки, поэтому для получения требуемого
    ; количества итераций необходимо выставить начальное значение индекса
    ; соотвествующим разности ближайшей сверху степени двойки и требуемого
    ; количества итераций.
    ; и учесть "неизбежность" первой итерации.
    ; M - CX 

    addl  @1, 2               ; R = M + 1 + 1
    subl  @1, @13             ; Index = R - CX

    ; Формируем значение теневых регистров

    slll  @1, 16              ; Index << 16
    or    @1, @4              ; 32[Index:0] | 32[0:Mask]

    ; Формируем значение регистра счетчика CX

    patch @1, 0
    setq  #CX, @1             ; CX = 64[ (32[Index:0] | 32[0:Mask]) : 0]

    ;=== Вычисляем тени индексных регистров ===
    ; Т.к. копирование побайтное, то аналогично счетчику,
    ; только теневой индекс должен инкрементироваться от нуля.

    rdl   mov.DesAddr
    rdl   mov.SrcAddr

;    rdl   chainop.STEP
;    mull  @10, @1

    patch @9,  @2              ; 64[ (32[0:0] | 32[0:Mask]) : DesAddr]
    patch @10, @2              ; 64[ (32[0:0] | 32[0:Mask]) : SrcAddr]
    setq  #DI, @2
    setq  #SI, @2

    ; Разрешаем теневые операции для выбранных регистров ("уходим в тень":-).

    getq  #MODR
    or    @1, MODR.SI | MODR.DI | MODR.CX
    setq  #MODR, @1

  complete
  movb.loop:

    ; Проверка условия выходи из цикла.

    exa   #CX
    je    @1, movb.epilogue
    jne   @2, movb.loop

    ; Тело цикла.

    rdb   #SI
    wrb   @1, #DI

  complete
  movb.epilogue:

    ; Строим возврат.

    rdl   mov.RetAddr
    jmp   @1

    getl  #BP
    rdl   mov.PrevBP
    setl  #BP, @1
    addl  @3, sizeof.ptr
    setl  #SP, @1

    ; Отменяем теневые операции ("выходим из тени":-)

    getq  #MODR
    and   @1, ~(MODR.SI | MODR.DI | MODR.CX)
    setq  #MODR, @1

  complete

  ;----------------------------------------------------------------------------

  movl:
  movl.prologue:

    jmp   movl.shadowing

    getl  #SP
    getl  #BP
    subl  @2, sizeof.ptr
    wrl   @2, #SP,-sizeof.ptr
    setl  #SP, @2
    setl  #BP, @3

  complete
  movl.shadowing:

    jmp   movl.loop

    ;=== Вычисляем тени регистра счетчика ===
    ; Вычисляем маску для заданного количества итераций,
    ; таким образом, что в ней выставлены все биты до бита представляющего
    ; наименьшую степень 2 не уступающую количеству итераций:

    rdl   mov.Counter
    subl  @1, 1               ; M - 1
    slrl  @1, 1
    or    @2, @1              ; M = M | (M >> 1)
    slrl  @1, 2
    or    @2, @1              ; M = M | (M >> 2)
    slrl  @1, 4
    or    @2, @1              ; M = M | (M >> 4)
    slrl  @1, 8
    or    @2, @1              ; M = M | (M >> 8)
    slrl  @1, 16
    or    @2, @1              ; Mask = M | (M >> 16)
    
    ; Расчитанная маска позволит выполнить количество итераций, соотвтетсвующее
    ; ближайшей сверху степени двойки, поэтому для получения требуемого
    ; количества итераций необходимо выставить начальное значение индекса
    ; соотвествующим разности ближайшей сверху степени двойки и требуемого
    ; количества итераций.
    ; и учесть "неизбежность" первой итерации.
    ; M - CX 

    addl  @1, 2               ; R = M + 1 + 1
    subl  @1, @13             ; Index = R - CX

    ; Формируем значение теневых регистров

    slll  @1, 16              ; Index << 16
    or    @1, @4              ; 32[Index:0] | 32[0:Mask]

    ; Формируем значение регистра счетчика CX

    patch @1, 0
    setq  #CX, @1             ; CX = 64[ (32[Index:0] | 32[0:Mask]) : 0]

    ;=== Вычисляем тени индексных регистров ===
    ; Т.к. копирование побайтное, то аналогично счетчику,
    ; только теневой индекс должен инкрементироваться от нуля.

    rdl   mov.DesAddr
    rdl   mov.SrcAddr
    slll  @9, 2               ; Mask << 2
    patch @1, @3              ; 64[ (32[0:0] | 32[0:Mask<<2]) : DesAddr]
    patch @2, @3              ; 64[ (32[0:0] | 32[0:Mask<<2]) : SrcAddr]
    setq  #DI, @2
    setq  #SI, @2

    ; Разрешаем теневые операции для выбранных регистров ("уходим в тень":-).

    getq  #MODR
    or    @1, MODR.SI | MODR.DI | MODR.CX
    setq  #MODR, @1

  complete
  movl.loop:

    ; Проверка условия выходи из цикла.

    exa   #CX
    je    @1, movl.epilogue
    jne   @2, movl.loop

    ; Тело цикла.

    rdl   #SI
    wrl   @1, #DI

  complete
  movl.epilogue:

    ; Строим возврат.

    rdl   mov.RetAddr
    jmp   @1

    getl  #BP
    rdl   mov.PrevBP
    setl  #BP, @1
    addl  @3, sizeof.ptr
    setl  #SP, @1

    ; Отменяем теневые операции ("выходим из тени":-)

    getq  #MODR
    and   @1, ~(MODR.SI | MODR.DI | MODR.CX)
    setq  #MODR, @1

  complete

  ;----------------------------------------------------------------------------

  movq:
  movq.prologue:

    jmp   movq.shadowing

    getl  #SP
    getl  #BP
    subl  @2, sizeof.ptr
    wrl   @2, #SP,-sizeof.ptr
    setl  #SP, @2
    setl  #BP, @3

  complete
  movq.shadowing:

    jmp   movq.loop

    ;=== Вычисляем тени регистра счетчика ===
    ; Вычисляем маску для заданного количества итераций,
    ; таким образом, что в ней выставлены все биты до бита представляющего
    ; наименьшую степень 2 не уступающую количеству итераций:

    rdl   mov.Counter
    subl  @1, 1               ; M - 1
    slrl  @1, 1
    or    @2, @1              ; M = M | (M >> 1)
    slrl  @1, 2
    or    @2, @1              ; M = M | (M >> 2)
    slrl  @1, 4
    or    @2, @1              ; M = M | (M >> 4)
    slrl  @1, 8
    or    @2, @1              ; M = M | (M >> 8)
    slrl  @1, 16
    or    @2, @1              ; Mask = M | (M >> 16)
    
    ; Расчитанная маска позволит выполнить количество итераций, соотвтетсвующее
    ; ближайшей сверху степени двойки, поэтому для получения требуемого
    ; количества итераций необходимо выставить начальное значение индекса
    ; соотвествующим разности ближайшей сверху степени двойки и требуемого
    ; количества итераций.
    ; и учесть "неизбежность" первой итерации.
    ; M - CX 

    addl  @1, 2               ; R = M + 1 + 1
    subl  @1, @13             ; Index = R - CX

    ; Формируем значение теневых регистров

    slll  @1, 16              ; Index << 16
    or    @1, @4              ; 32[Index:0] | 32[0:Mask]

    ; Формируем значение регистра счетчика CX

    patch @1, 0
    setq  #CX, @1             ; CX = 64[ (32[Index:0] | 32[0:Mask]) : 0]

    ;=== Вычисляем тени индексных регистров ===
    ; Т.к. копирование побайтное, то аналогично счетчику,
    ; только теневой индекс должен инкрементироваться от нуля.

    rdl   mov.DesAddr
    rdl   mov.SrcAddr
    slll  @9, 3               ; Mask << 3
    patch @1, @3              ; 64[ (32[0:0] | 32[0:Mask<<3]) : DesAddr]
    patch @2, @3              ; 64[ (32[0:0] | 32[0:Mask<<3]) : SrcAddr]
    setq  #DI, @2
    setq  #SI, @2

    ; Разрешаем теневые операции для выбранных регистров ("уходим в тень":-).

    getq  #MODR
    or    @1, MODR.SI | MODR.DI | MODR.CX
    setq  #MODR, @1

  complete
  movq.loop:

    ; Проверка условия выходи из цикла.

    exa   #CX
    je    @1, movq.epilogue
    jne   @2, movq.loop

    ; Тело цикла.

    rdq   #SI
    wrq   @1, #DI

  complete
  movq.epilogue:

    ; Строим возврат.

    rdl   mov.RetAddr
    jmp   @1

    getl  #BP
    rdl   mov.PrevBP
    setl  #BP, @1
    addl  @3, sizeof.ptr
    setl  #SP, @1

    ; Отменяем теневые операции ("выходим из тени":-)

    getq  #MODR
    and   @1, ~(MODR.SI | MODR.DI | MODR.CX)
    setq  #MODR, @1

  complete

.end
