#include <HDL51001_ccf.h>
#include <HDL50001_gpio.h>
#include <wdt.h>
#include <uart.h>
#include "FreeRTOS.h"
#include "task.h"
#include "croutine.h"
#include "queue.h"

xQueueHandle xQueue;

void vSenderTask(void *pvParameters)
{
     int lValueToSend;

     portBASE_TYPE xStatus;
     lValueToSend = (int) pvParameters;

     for (;;)
     {
         xStatus = xQueueSendToBack(xQueue, &lValueToSend, 0);
         if (xStatus != pdPASS)
         {
             uart_send_str("Could not send to the queue.\r\n", UART0);
         }

         taskYIELD();
     }
}

void vReceiverTask(void *pvParameters)
{
     int lReceivedValue;
     portBASE_TYPE xStatus;

     for (;;)
     {
         if (uxQueueMessagesWaiting(xQueue) != 0)
         {
             uart_send_str("Queue should have been empty!\r\n", UART0);
         }

         xStatus = xQueueReceive(xQueue, &lReceivedValue, 100);

         if (xStatus == pdPASS)
         {
             uart_send_int(lReceivedValue, UART0);
         }
         else
         {
             uart_send_str("Could not receive from the queue.\r\n", UART0);
         }
     }
}


int main(void)
{

     uart();

     xQueue = xQueueCreate(5, sizeof(int));

     if (xQueue != NULL)
     {
         xTaskCreate(vSenderTask,(signed char*) "Sender1", 
			configMINIMAL_STACK_SIZE, (void *) 100, 1, NULL);
         xTaskCreate(vSenderTask,(signed char*) "Sender2", 
			configMINIMAL_STACK_SIZE, (void *) 200, 1, NULL);
         xTaskCreate(vReceiverTask,(signed char*) "Receiver", 
			configMINIMAL_STACK_SIZE, (void *)0xdead, 2, NULL);

         vTaskStartScheduler();
     }
     else
     {
     /* Если очередь не создана */
     }


}


void uart()
{
     UART_InitTypeDef UART_InitStructure;
     UART_InitStructure.BaudRate = 38400; //set baudrate
     UART_InitStructure.TypeParity = 0x00000000; //parity control type
     UART_InitStructure.Parity = 0x00000000; //enable parity control
     UART_InitStructure.FlowControl = 0x00000000; //enable cts/rts
     UART_InitStructure.Mode = 0x00000003; //rx enable - 1 bit, tx enable - 2 bit (rx + tx en)

     GPIOB->BPS = 0x00000300;         //alternative port function for uart0
     uart_init(UART0, &UART_InitStructure);
}

//void vApplicationIdleHook( void )
//{
//    vCoRoutineSchedule();
//}


